-- KillCounterProcess.lua
--
-- Event and kill processing
--
-- Globals: KillCounter
-- Globals: bit, UnitLevel, time
-- Globals: COMBATLOG_OBJECT_AFFILIATION_MASK
-- Globals: COMBATLOG_OBJECT_CONTROL_MASK
-- Globals: COMBATLOG_OBJECT_TYPE_MASK
-- Globals: COMBATLOG_OBJECT_AFFILIATION_MINE
-- Globals: COMBATLOG_OBJECT_AFFILIATION_PARTY
-- Globals: COMBATLOG_OBJECT_AFFILIATION_RAID
-- Globals: COMBATLOG_OBJECT_CONTROL_PLAYER
-- Globals: COMBATLOG_OBJECT_TYPE_PLAYER
-- Globals: COMBATLOG_OBJECT_TYPE_NPC
-- Globals: COMBATLOG_OBJECT_TYPE_PET
-- Globals: COMBATLOG_OBJECT_TYPE_GUARDIAN


local KC = KillCounter;
local debugEnabled;
KC:EnableDebug(function(e) debugEnabled = e; end);

local select = select;

local band = bit.band;
local bor = bit.bor;

local INTERESTING_TYPE_MASK =
    bor(COMBATLOG_OBJECT_AFFILIATION_MASK,
        COMBATLOG_OBJECT_CONTROL_MASK,
        COMBATLOG_OBJECT_TYPE_MASK);

local ASSOC_BITS_ME  = 0x03;
local ASSOC_BITS_PET = 0x02;

local ASSOC_BITS_PARTY = 0x0c;
local ASSOC_BITS_RAID  = 0x08;

local ASSOC_BITS_OTHER_PLAYER = 0x10;
local ASSOC_BITS_OTHER_NPC    = 0x20;

local MAX_ACCUMULATION_TIME = 180; -- 3 minutes should be plenty
local CLEANUP_SWEEP_INTERVAL = 60; -- Cleanup every 60 seconds

local NEXT_CLEANUP_SWEEP = 0;
local ACCUMULATED_CONTRIBUTIONS = {};
local TIMESTAMP_CONTRIBUTIONS = {};

local function CheckCleanup()
    local now = time();
    if (now < NEXT_CLEANUP_SWEEP) then
        return;
    end

    if (debugEnabled) then
        KC:Debug("Doing cleanup sweep...");
    end

    local cutoff = now - MAX_ACCUMULATION_TIME;
    NEXT_CLEANUP_SWEEP = now + CLEANUP_SWEEP_INTERVAL;

    for guid, lastTime in pairs(TIMESTAMP_CONTRIBUTIONS) do
        if (lastTime < cutoff) then
            if (debugEnabled) then
                KC:Debug("Retiring stale GUID", guid);
            end
            TIMESTAMP_CONTRIBUTIONS[guid] = nil;
            ACCUMULATED_CONTRIBUTIONS[guid] = nil;
        end
    end
end

local function KC_ClassifyType(flags)
    local mFlags = band(flags, INTERESTING_TYPE_MASK);

    local ret;

    if (band(mFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) ~= 0) then
        ret = "Mine";
    elseif (band(mFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY) ~= 0) then
        ret = "Party";
    elseif (band(mFlags, COMBATLOG_OBJECT_AFFILIATION_RAID) ~= 0) then
        ret = "Raid";
    else
        ret = "Other";
    end

    if (band(mFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) ~= 0) then
        ret = ret .. "Player";
    else
        ret = ret .. "Npc";
    end

    if (band(mFlags, COMBATLOG_OBJECT_TYPE_PLAYER) ~= 0) then
        ret = ret .. "Player";
    elseif (band(mFlags, COMBATLOG_OBJECT_TYPE_NPC) ~= 0) then
        ret = ret .. "Npc";
    elseif (band(mFlags, COMBATLOG_OBJECT_TYPE_PET) ~= 0) then
        ret = ret .. "Pet";
    elseif (band(mFlags, COMBATLOG_OBJECT_TYPE_GUARDIAN) ~= 0) then
        ret = ret .. "Guardian";
    else
        ret = ret .. "Object";
    end

    return ret;
end

local function KC_DecodeAssociationValue(assoc)
    local ret;
    if (band(assoc, ASSOC_BITS_ME) == ASSOC_BITS_ME) then
        ret = ""; -- Default to me
    elseif (band(assoc, ASSOC_BITS_PET) ~= 0) then
        ret = "P"; -- Pet only
    else
        ret = "X"; -- I wasn't involved
    end

    if (band(assoc, ASSOC_BITS_PARTY) == ASSOC_BITS_PARTY) then
        ret = ret .. "G"; -- Group
    elseif (band(assoc, ASSOC_BITS_RAID) ~= 0) then
        ret = ret .. "R"; -- Raid
    end

    if (band(assoc, ASSOC_BITS_OTHER_PLAYER) ~= 0) then
        ret = ret .. "O"; -- Other
    end

    if (band(assoc, ASSOC_BITS_OTHER_NPC) ~= 0) then
        ret = ret .. "N"; -- NPC
    end

    return ret;
end

local CODE_LIST = {};
for _, CC in pairs({"", "O", "N", "ON"}) do
    for _, CB in pairs({"", "G", "R"}) do
        for _, CA in pairs({"","P","X"}) do
            table.insert(CODE_LIST, CA .. CB .. CC);
        end
    end
end
table.insert(CODE_LIST, "I");
KC.CODE_LIST = CODE_LIST;

local function KC_CalculateAssociationValue(flags)
    -- If it's me, this is pretty easy
    if (band(flags, COMBATLOG_OBJECT_AFFILIATION_MINE) ~= 0) then
        if (band(flags, COMBATLOG_OBJECT_TYPE_PLAYER) ~= 0) then
            return ASSOC_BITS_ME;
        end
        return ASSOC_BITS_PET;
    end

    if (band(flags, COMBATLOG_OBJECT_AFFILIATION_PARTY) ~= 0) then
        return ASSOC_BITS_PARTY;
    elseif (band(flags, COMBATLOG_OBJECT_AFFILIATION_RAID) ~= 0) then
        return ASSOC_BITS_RAID;
    end

    if (band(flags, COMBATLOG_OBJECT_CONTROL_PLAYER) ~= 0) then
        return ASSOC_BITS_OTHER_PLAYER;
    end
    return ASSOC_BITS_OTHER_NPC;
end

local function KC_PopulateAssociationValue(t, flags)
    if (type(flags) ~= "number") then
        return nil;
    end
    if (band(flags, INTERESTING_TYPE_MASK) ~= flags) then
        return nil;
    end
    local ret = KC_CalculateAssociationValue(flags);
    t[flags] = ret;
    return ret;
end

local ASSOCIATION_LOOKUP = {};
setmetatable(ASSOCIATION_LOOKUP,
             {__index = KC_PopulateAssociationValue});

local function KC_GetAssociationValue(flags)
    return ASSOCIATION_LOOKUP[band(flags, INTERESTING_TYPE_MASK)];
end

local function KC_OnCombatDamage(timestamp,
                                  sGUID, sName, sFlags,
                                  dGUID, dName, dFlags,
                                  amount, ...)
    CheckCleanup();
    local cur = ACCUMULATED_CONTRIBUTIONS[dGUID] or 0;
    local sAssoc = KC_GetAssociationValue(sFlags);
    local new = bor(cur, sAssoc);
    ACCUMULATED_CONTRIBUTIONS[dGUID] = new;
    TIMESTAMP_CONTRIBUTIONS[dGUID] = timestamp;

    -- sAssoc = KC_DecodeAssociationValue(new);
    -- print("DAMAGE ", dGUID, " ", dName, sAssoc);
end

local function KC_RecordDeath(guid, name, flags, assoc)
    local handle = KC:GetHandle(guid);
    if (handle == nil) then
        return;
    end
    local record = KC:GetRecord(handle, true, name);
    --print("DEATH", guid, name, KC_ClassifyType(flags), assoc);
    record:AddKill(assoc, UnitLevel("player"));
    record:Free();
    KC:RingBell(1);
end

local function KC_OnCombatDeath(timestamp, isDeath, dGUID, dName, dFlags)
    local cur = ACCUMULATED_CONTRIBUTIONS[dGUID];
    if (cur ~= nil) then
        ACCUMULATED_CONTRIBUTIONS[dGUID] = nil;
        TIMESTAMP_CONTRIBUTIONS[dGUID] = nil;
    else
        cur = 0;
    end

    if (not isDeath) then
        -- This wasn't a real death, we can ignore it for now
        return;
    end

    local assoc = KC_DecodeAssociationValue(cur);
    if ((assoc == "X") or (assoc == "XN")
        or (assoc == "XO") or (assoc == "XON")) then
        -- I wasn't involved, this is boring
        return;
    end

    KC_RecordDeath(dGUID, dName, dFlags, assoc);
end

function KC:OnCombatLogEvent(timestamp, event, sGUID, sName, sFlags,
                                    dGUID, dName, dFlags, ...)
    if (event == "UNIT_DIED") then
        return KC_OnCombatDeath(timestamp, true, dGUID, dName, dFlags);
    end

    if (event == "UNIT_DESTROYED") then
        return KC_OnCombatDeath(timestamp, false, dGUID, dName, dFlags);
    end

    if (event == "SWING_DAMAGE") then
        KC_OnCombatDamage(timestamp,
                           sGUID, sName, sFlags, dGUID, dName, dFlags,
                           ...);
        return
    end

    if ((event == "SPELL_DAMAGE")
        or (event == "RANGE_DAMAGE")
        or (event == "SPELL_PERIODIC_DAMAGE")
            or (event == "DAMAGE_SHIELD")
            or (event == "DAMAGE_SPLIT")) then
        KC_OnCombatDamage(timestamp,
                           sGUID, sName, sFlags, dGUID, dName, dFlags,
                           select(4, ...));
        return
    end

    if ((event == "SPELL_DRAIN")
        or (event == "SPELL_LEECH")
        or (event == "SPELL_PERIODIC_DRAIN")
        or (event == "SPELL_PERIODIC_LEECH")
        or (event == "RANGE_DRAIN")
        or (event == "RANGE_LEECH")) then
        local _,_,_,amount, powerType, extraAmount = ...;
        if (powerType == -2) then
            KC_OnCombatDamage(timestamp,
                               sGUID, sName, sFlags, dGUID, dName, dFlags,
                               amount, powerType, extraAmount);
        end
        return
    end
end
